/*
*         OpenPBS (Portable Batch System) v2.3 Software License
* 
* Copyright (c) 1999-2000 Veridian Information Solutions, Inc.
* All rights reserved.
* 
* ---------------------------------------------------------------------------
* For a license to use or redistribute the OpenPBS software under conditions
* other than those described below, or to purchase support for this software,
* please contact Veridian Systems, PBS Products Department ("Licensor") at:
* 
*    www.OpenPBS.org  +1 650 967-4675                  sales@OpenPBS.org
*                        877 902-4PBS (US toll-free)
* ---------------------------------------------------------------------------
* 
* This license covers use of the OpenPBS v2.3 software (the "Software") at
* your site or location, and, for certain users, redistribution of the
* Software to other sites and locations.  Use and redistribution of
* OpenPBS v2.3 in source and binary forms, with or without modification,
* are permitted provided that all of the following conditions are met.
* After December 31, 2001, only conditions 3-6 must be met:
* 
* 1. Commercial and/or non-commercial use of the Software is permitted
*    provided a current software registration is on file at www.OpenPBS.org.
*    If use of this software contributes to a publication, product, or
*    service, proper attribution must be given; see www.OpenPBS.org/credit.html
* 
* 2. Redistribution in any form is only permitted for non-commercial,
*    non-profit purposes.  There can be no charge for the Software or any
*    software incorporating the Software.  Further, there can be no
*    expectation of revenue generated as a consequence of redistributing
*    the Software.
* 
* 3. Any Redistribution of source code must retain the above copyright notice
*    and the acknowledgment contained in paragraph 6, this list of conditions
*    and the disclaimer contained in paragraph 7.
* 
* 4. Any Redistribution in binary form must reproduce the above copyright
*    notice and the acknowledgment contained in paragraph 6, this list of
*    conditions and the disclaimer contained in paragraph 7 in the
*    documentation and/or other materials provided with the distribution.
* 
* 5. Redistributions in any form must be accompanied by information on how to
*    obtain complete source code for the OpenPBS software and any
*    modifications and/or additions to the OpenPBS software.  The source code
*    must either be included in the distribution or be available for no more
*    than the cost of distribution plus a nominal fee, and all modifications
*    and additions to the Software must be freely redistributable by any party
*    (including Licensor) without restriction.
* 
* 6. All advertising materials mentioning features or use of the Software must
*    display the following acknowledgment:
* 
*     "This product includes software developed by NASA Ames Research Center,
*     Lawrence Livermore National Laboratory, and Veridian Information 
*     Solutions, Inc.
*     Visit www.OpenPBS.org for OpenPBS software support,
*     products, and information."
* 
* 7. DISCLAIMER OF WARRANTY
* 
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND. ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT
* ARE EXPRESSLY DISCLAIMED.
* 
* IN NO EVENT SHALL VERIDIAN CORPORATION, ITS AFFILIATED COMPANIES, OR THE
* U.S. GOVERNMENT OR ANY OF ITS AGENCIES BE LIABLE FOR ANY DIRECT OR INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
* 
* This license will be governed by the laws of the Commonwealth of Virginia,
* without reference to its choice of law rules.
*/

/*
 * dis_read.c - contains function to read and decode the DIS
 *	encoded requests and replies.
 */

#include <pbs_config.h>   /* the master config generated by configure */

#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include "dis.h"
#include "libpbs.h"
#include "list_link.h"
#include "server_limits.h"
#include "attribute.h"
#include "log.h"
#include "pbs_error.h"
#include "credential.h"
#include "batch_request.h"
#include "net_connect.h"



/* External Global Data */

extern char	*msg_nosupport;
extern char     *PBatchReqType[];

extern int       LOGLEVEL;


/*
 * dis_request_read - read in an DIS encoded request from the network 
 * and decodes it:
 *	Read and decode the request into the request structures 
 *
 *	returns: 0 if request read ok, batch_request pointed to by request is
 *		   updated.
 *		-1 if EOF (no request but no error)
 *		>0 if errors ( a PBSE_ number)
 */

int dis_request_read(

  int                   sfds,	 /* socket to read */
  struct batch_request *request) /* server internal structure */

  {
  char *id = "dis_request_read";

  int	 proto_type;
  int	 proto_ver;
  int	 rc; 	/* return code */

  DIS_tcp_setup(sfds);	/* setup for DIS over tcp */

  /* Decode the Request Header, that will tell the request type */

  if ((rc = decode_DIS_ReqHdr(sfds,request,&proto_type,&proto_ver))) 
    {
    if (rc == DIS_EOF)
      {
      return(EOF);
      }

    sprintf(log_buffer,"req header bad, dis error %d", 
      rc);

    LOG_EVENT(PBSEVENT_DEBUG,PBS_EVENTCLASS_REQUEST,"?",
      log_buffer);

    return(PBSE_DISPROTO);
    }

  if (proto_ver != PBS_BATCH_PROT_VER) 
    {
    sprintf(log_buffer,"conflicting version numbers, %d detected, %d expected",
      proto_ver,
      PBS_BATCH_PROT_VER);

    LOG_EVENT(PBSEVENT_DEBUG,PBS_EVENTCLASS_REQUEST,"?",
      log_buffer);

    return(PBSE_DISPROTO);
    }

  if ((request->rq_type < 0) || (request->rq_type > PBS_BATCH_Disconnect))
    {
    sprintf(log_buffer,"invalid request type: %d",
      request->rq_type);

    LOG_EVENT(PBSEVENT_DEBUG,PBS_EVENTCLASS_REQUEST,"?",
      log_buffer);

    return(PBSE_DISPROTO);
    }

  /* Decode the Request Body based on the type */

  if (LOGLEVEL >= 5)
    {
    sprintf(log_buffer,"decoding command %s from %s",
      PBatchReqType[request->rq_type],
      request->rq_user);

    LOG_EVENT(
      PBSEVENT_DEBUG,
      PBS_EVENTCLASS_REQUEST,
      id,
      log_buffer);
    }

  switch (request->rq_type) 
    {
    case PBS_BATCH_Disconnect:
    
      return(-1);  /* set EOF return */

      /*NOTREACHED*/

      break;

    case PBS_BATCH_QueueJob:

      CLEAR_HEAD(request->rq_ind.rq_queuejob.rq_attr);

      rc = decode_DIS_QueueJob(sfds,request);

      break;

    case PBS_BATCH_JobCred:

      rc = decode_DIS_JobCred(sfds,request);

      break;

    case PBS_BATCH_jobscript:
    case PBS_BATCH_MvJobFile:

      rc = decode_DIS_JobFile(sfds,request);

      break;

    case PBS_BATCH_RdytoCommit:
    case PBS_BATCH_Commit:
    case PBS_BATCH_Rerun:

      rc = decode_DIS_JobId(sfds,request->rq_ind.rq_commit);

      break;

    case PBS_BATCH_DeleteJob:
    case PBS_BATCH_HoldJob:
    case PBS_BATCH_ModifyJob:

      rc = decode_DIS_Manage(sfds, request);

      break;

    case PBS_BATCH_MessJob:

      rc = decode_DIS_MessageJob(sfds, request);

      break;

    case PBS_BATCH_Shutdown:

      rc = decode_DIS_ShutDown(sfds,request);

      break;

    case PBS_BATCH_SignalJob:

      rc = decode_DIS_SignalJob(sfds,request);

      break;

    case PBS_BATCH_StatusJob:

      rc = decode_DIS_Status(sfds,request);

      break;

#ifndef PBS_MOM

    case PBS_BATCH_LocateJob:

      rc = decode_DIS_JobId(sfds,request->rq_ind.rq_locate);

      break;

    case PBS_BATCH_Manager:
    case PBS_BATCH_ReleaseJob:

      rc = decode_DIS_Manage(sfds,request);

      break;

    case PBS_BATCH_MoveJob:
    case PBS_BATCH_OrderJob:

      rc = decode_DIS_MoveJob(sfds,request);

      break;

    case PBS_BATCH_RunJob:
    case PBS_BATCH_AsyrunJob:
    case PBS_BATCH_StageIn:

      rc = decode_DIS_RunJob(sfds,request);

      break;

    case PBS_BATCH_SelectJobs:
    case PBS_BATCH_SelStat:

      CLEAR_HEAD(request->rq_ind.rq_select);

      rc = decode_DIS_svrattrl(sfds,&request->rq_ind.rq_select);

      break;

    case PBS_BATCH_StatusNode:
    case PBS_BATCH_StatusQue:
    case PBS_BATCH_StatusSvr:
    
      rc = decode_DIS_Status(sfds,request);

      break;

    case PBS_BATCH_TrackJob:

      rc = decode_DIS_TrackJob(sfds,request);

      break;

    case PBS_BATCH_Rescq:
    case PBS_BATCH_ReserveResc:
    case PBS_BATCH_ReleaseResc:

      rc = decode_DIS_Rescl(sfds,request);

      break;

    case PBS_BATCH_RegistDep:

      rc = decode_DIS_Register(sfds,request);

      break;

    case PBS_BATCH_AuthenUser:

      rc = decode_DIS_Authen(sfds,request);

      break;

    case PBS_BATCH_JobObit:

      rc = decode_DIS_JobObit(sfds,request);

      break;

#else  /* PBS_MOM */
	
    /* pbs_mom services */
	
    case PBS_BATCH_CopyFiles:
    case PBS_BATCH_DelFiles:

      rc = decode_DIS_CopyFiles(sfds,request);

      break;

#endif	/* PBS_MOM */

    default:

      sprintf(log_buffer,"%s: %d from %s", 
        msg_nosupport, 
        request->rq_type, 
        request->rq_user);

      LOG_EVENT(
        PBSEVENT_DEBUG, 
        PBS_EVENTCLASS_REQUEST, 
        id,
        log_buffer);

      rc = PBSE_UNKREQ;

      break;
    }  /* END switch (request->rq_type) */
		
  if (rc == 0) 
    {
    /* Decode the Request Extension, if present */

    if ((rc = decode_DIS_ReqExtend(sfds,request))) 
      {  
      sprintf(log_buffer,"req extension bad, dis error %d (%s), type=%s", 
        rc,
        dis_emsg[rc],
        PBatchReqType[request->rq_type]);

      LOG_EVENT(
        PBSEVENT_DEBUG, 
        PBS_EVENTCLASS_REQUEST, 
        "?", 
        log_buffer);

      rc = PBSE_DISPROTO;
      }
    } 
  else if (rc != PBSE_UNKREQ) 
    {
    sprintf(log_buffer,"req body bad, dis error %d (%s), type=%s",
      rc, 
      dis_emsg[rc],
      PBatchReqType[request->rq_type]);

    LOG_EVENT(PBSEVENT_DEBUG,PBS_EVENTCLASS_REQUEST,"?",log_buffer);

    rc = PBSE_DISPROTO;
    }

  return(rc);
  }  /* END dis_request_read() */






/*
 * DIS_reply_read - read and decode DIS based batch reply
 *
 *	Returns  0 - reply read and decoded ok
 *		!0 - otherwise
 */


int DIS_reply_read(

  int                 sock,
  struct batch_reply *preply)

  {
  DIS_tcp_setup(sock);

  return(decode_DIS_replySvr(sock,preply));
  }

/* END dis_read.c */


